@@ -1,8 +1,10 @@
 """System tests for Session."""
 
 # SPDX-License-Identifier: Apache-2.0
+# Copyright Tumult Labs 2022
 
 import datetime
+import math
 from typing import Any, Dict, List, Mapping, Optional, Tuple, Type, Union
 from unittest.mock import patch
 
@@ -11,6 +13,7 @@ from parameterized import parameterized
 from pyspark.sql import DataFrame, SparkSession
 from pyspark.sql.types import LongType, StringType, StructField, StructType
 
+from tmlt.analytics._noise_info import _NoiseMechanism
 from tmlt.analytics._schema import (
     ColumnDescriptor,
     ColumnType,
@@ -24,10 +27,12 @@ from tmlt.analytics.privacy_budget import PrivacyBudget, PureDPBudget, RhoZCDPBu
 from tmlt.analytics.query_builder import QueryBuilder
 from tmlt.analytics.query_expr import (
     AnalyticsDefault,
+    AverageMechanism,
     CountDistinctMechanism,
     CountMechanism,
     Filter,
     FlatMap,
+    GroupByBoundedAverage,
     GroupByBoundedSTDEV,
     GroupByBoundedSum,
     GroupByCount,
@@ -54,6 +59,7 @@ from tmlt.core.measurements.interactive_measurements import (
 from tmlt.core.measures import PureDP, RhoZCDP
 from tmlt.core.metrics import DictMetric, SymmetricDifference
 from tmlt.core.utils.exact_number import ExactNumber
+from tmlt.core.utils.parameters import calculate_noise_scale
 from tmlt.core.utils.testing import PySparkTest
 
 # Shorthands for some values used in tests
@@ -719,6 +725,78 @@ class TestEvaluate(PySparkTest):
         )
         self.assert_frame_equal_with_sort(actual_sdf.toPandas(), expected_df)
 
+    @parameterized.expand(
+        [
+            (
+                GroupByCount(
+                    child=PrivateSource("private"),
+                    groupby_keys=KeySet.from_dict({}),
+                    mechanism=CountMechanism.LAPLACE,
+                ),
+                PureDPBudget(11),
+                PureDPBudget(7),
+                [
+                    {
+                        "noise_mechanism": _NoiseMechanism.GEOMETRIC,
+                        "noise_parameter": (1.0 / 7.0),
+                    }
+                ],
+            ),
+            (
+                GroupByBoundedAverage(
+                    child=PrivateSource("private"),
+                    groupby_keys=KeySet.from_dict({}),
+                    low=-111,
+                    high=234,
+                    mechanism=AverageMechanism.GAUSSIAN,
+                    measure_column="X",
+                ),
+                RhoZCDPBudget(31),
+                RhoZCDPBudget(11),
+                [
+                    # Noise for the sum query (which uses half the budget)
+                    {
+                        "noise_mechanism": _NoiseMechanism.DISCRETE_GAUSSIAN,
+                        # the upper and lower bounds of the sum aggregation
+                        # are -173 and 172;
+                        # this is (lower - midpoint) and (upper-midpoint) respectively
+                        "noise_parameter": (
+                            calculate_noise_scale(
+                                173, ExactNumber(11) / ExactNumber(2), RhoZCDP()
+                            )
+                            ** 2
+                        ).to_float(round_up=False),
+                    },
+                    # Noise for the count query (which uses half the budget)
+                    {
+                        "noise_mechanism": _NoiseMechanism.DISCRETE_GAUSSIAN,
+                        "noise_parameter": (
+                            calculate_noise_scale(
+                                1, ExactNumber(11) / ExactNumber(2), RhoZCDP()
+                            )
+                            ** 2
+                        ).to_float(round_up=False),
+                    },
+                ],
+            ),
+        ]
+    )
+    def test_noise_info(
+        self,
+        query_expr: QueryExpr,
+        session_budget: PrivacyBudget,
+        query_budget: PrivacyBudget,
+        expected: List[Dict[str, Any]],
+    ):
+        """Test _noise_info."""
+        session = Session.from_dataframe(
+            privacy_budget=session_budget, source_id="private", dataframe=self.sdf
+        )
+        # pylint: disable=protected-access
+        info = session._noise_info(query_expr, query_budget)
+        # pylint: enable=protected-access
+        self.assertEqual(info, expected)
+
     @parameterized.expand(
         [(PureDPBudget(float("inf")),), (RhoZCDPBudget(float("inf")),)]
     )
@@ -1209,6 +1287,25 @@ class TestEvaluate(PySparkTest):
 
     @parameterized.expand([(PureDPBudget(20),), (RhoZCDPBudget(20),)])
     def test_partition_on_grouping_column(self, budget: PrivacyBudget):
+        """Tests that you can partition on grouping columns."""
+        grouping_df = self.spark.createDataFrame(pd.DataFrame({"new": [1, 2]}))
+        session = Session.from_dataframe(
+            privacy_budget=budget,
+            source_id="private",
+            dataframe=self.sdf.crossJoin(grouping_df),
+            grouping_column="new",
+        )
+        new_sessions = session.partition_and_create(
+            source_id="private",
+            privacy_budget=budget,
+            attr_name="new",
+            splits={"new1": 1, "new2": 2},
+        )
+        new_sessions["new1"].evaluate(QueryBuilder("new1").count(), budget)
+        new_sessions["new2"].evaluate(QueryBuilder("new2").count(), budget)
+
+    @parameterized.expand([(PureDPBudget(20),), (RhoZCDPBudget(20),)])
+    def test_partition_on_flatmap_grouping_column(self, budget: PrivacyBudget):
         """Tests that you can partition on columns created by grouping flat maps."""
         session = Session.from_dataframe(
             privacy_budget=budget, source_id="private", dataframe=self.sdf
@@ -1407,6 +1504,23 @@ class TestEvaluate(PySparkTest):
             len(list(self.spark.sparkContext._jsc.sc().getRDDStorageInfo())), 0
         )
 
+    def test_grouping_noninteger_stability(self) -> None:
+        """Test that zCDP grouping_column and non-integer stabilities work."""
+        grouped_df = self.spark.createDataFrame(
+            pd.DataFrame({"id": [7, 7, 8, 9], "group": [0, 1, 0, 1]})
+        )
+        ks = KeySet.from_dict({"group": [0, 1]})
+        query = QueryBuilder("id").groupby(ks).count()
+
+        sess = Session.from_dataframe(
+            RhoZCDPBudget(float("inf")),
+            "id",
+            grouped_df,
+            stability=math.sqrt(2),
+            grouping_column="group",
+        )
+        sess.evaluate(query, RhoZCDPBudget(1))
+
 
 class TestInvalidSession(PySparkTest):
     """Tests for invalid session."""
